//
//  File  : nbody.cpp
//  (c)HOOPOE ProGroup 2013
//

//
// C= Programming Sample
//

/// Copyright string
char* copyright = "(c)HOOPOE ProGroup 2013";

#include <stdio.h>
#include <C=.h>

#include <stdlib.h>
#include <math.h>

#include <intrin.h>
#pragma intrinsic(__rdtsc)

__int64 time;

/// data

const int N=1024 * 1024;  /// total bodies
const int M=1024 * 64;    /// max bodies to fit GPU memory
///const int N=1024 * 64; /// total bodies
///const int M=1024 * 32; /// max bodies to fit GPU memory
///const int N=1024 * 8;  /// total bodies
///const int M=1024 * 2;  /// max bodies to fit GPU memory

float pos_par[N * 4];

float vel_par[N * 4];

float force_par[N / M][N * 3];

int bodies = N;
 
float soften = .00125f;
float pscale = 1.54f;
float vscale = 8.0f;
float damp = 0.995f;
float step = 0.016f;

/// auxilliary functions

float dot(float v0[3], float v1[3])
{
    return v0[0] * v1[0] + v0[1] * v1[1] + v0[2] * v1[2];
}

/// initialize N-bodies

void init_bodies(float* pos, float* vel, float pscale, float vscale, int bodies)
{
    float scale = pscale * bodies / 1024.f;

    if(scale < 1.0f)
    {
        scale = pscale;
    }
    vscale *= scale;

    int p = 0, v = 0;

    for(int i = 0; i < bodies;) 
    {
        float point[3];

        point[0] = rand() / (float)RAND_MAX * 2 - 1;
        point[1] = rand() / (float)RAND_MAX * 2 - 1;
        point[2] = rand() / (float)RAND_MAX * 2 - 1;

        float len = dot(point, point);
        
        if (len > 1)
        {
            continue;
        }

        pos[p++] = point[0] * scale; /// X
        pos[p++] = point[1] * scale; /// Y
        pos[p++] = point[2] * scale; /// Z
        pos[p++] = 1.0f; // mass
        vel[v++] = point[0] * vscale; /// X-velocity
        vel[v++] = point[1] * vscale; /// Y-velocity
        vel[v++] = point[2] * vscale; /// Z-velocity
        v++;

        i++;
    }
}

/// compute N-bodies

void compute_bodies(float step, float* pos, float* vel, 
                    float force[N / M][N * 3])
{
    /// convert the time step to float (it is passed as double by the compiler)
    float delta = (float)step;

    /// decompose forces
    parallel(N / M)
    {
        int start = CPX_RANGE_START(bodies);
        int* shared startp = &start;

        /// decompose positions
        parallel(N / M)
        {
            for(;;)
            {
                int start = CPX_RANGE_START(bodies);
                int len = CPX_RANGE_END(bodies) - start;

                /// access through pointers
                float* shared velp = &vel[start * 4];
                float* shared posp = &pos[start * 4];
                float* shared forp = &force[pix()][*startp * 3];
                int* shared lenp = &len;

                /// error-resilient programming facility
                int error = 0;
                /// access through pointer
                int* shared errp = &error;

                /// use parallel(1) to send everything to remote hosts 
                /// if available
                parallel(1)
                {
                    coload(
                           /// input positions
                           cpx_in (float*, posp, len * sizeof(float) * 4),
                           /// output forces
                           cpx_out(float*, forp, len * sizeof(float) * 3),

                           /// input constants
                           cpx_const(int, len), 
                           cpx_const(float, damp), 
                           cpx_const(float, soften),
                           /// error pointer
                           cpx_error(errp))
                    {
                        /// define rsqrtf for CPU to match CUDA's definition
                        /// change it to rsqrt to match OpenCL's definition
                        #define rsqrtf(x) (1.0f / sqrtf(x))

                        /// use parallel(1) to send everything to GPU 
                        /// if available
                        parallel(1)
                        {
                            /// set up memory local to GPU threads 
                            /// we intend to use 256 threads per 
                            /// execution block
                            float cache[256 * 4];
                            
                            /// co-load all GPUs 
                            /// (and describe all dependencies)
                            coload(
                                   /// input positions
                                   cpx_in (float*, posp, 
                                           len * sizeof(float) * 4),
                                   /// output forces
                                   cpx_out(float*, forp, 
                                           len * sizeof(float) * 3),

                                   /// input constants
                                   cpx_const(int, len), 
                                   cpx_const(float, damp), 
                                   cpx_const(float, soften),

                                   /// specify the size of the memory buffer 
                                   /// shared by GPU threads 
                                   /// running on the same multiprocessor
                                   cpx_cache(sizeof(cache)),
                                   /// requested number of threads 
                                   /// to run on a multiprocessor
                                   /// (threads per thread block)
                                   cpx_warp(256),

                                   /// the actual number of iterations 
                                   /// to perform
                                   cpx_range(len))

                                   /// NOTE: ((cpx_range / pno()) % cpx_warp) 
                                   ///         must be 0
                            {
                                /// loop over bodies 
                                /// (use built-in decomposition macros)
                                for (int i = CPX_RANGE_START(len), 
                                       end = CPX_RANGE_END(len) ; i < end; i++)
                                {
                                    /// compute array indices
                                    int idx = 4 * i;
                                    int fdx = 3 * i;

                                    /// temporary accumulator
                                    float acc[3] = {0, 0, 0};

                                    /// work in 256-body blocks
                                    for(int j = 0; j < len; j += 256) 
                                    {
                                        /// fill in a shared memory buffer 
                                        /// (256 interations on CPU, 
                                        ///  1 iteration on GPU)
                                        for(int k = i; k < ((end - i) < 256 ? 
                                                       end : (i + 256)); k++)
                                        {
                                            int jdx = 4 * (j + (k % 256));
                                            int kdx = 4 * (k % 256);

                                            /// copy k-th element from 
                                            /// the current 256-body block to 
                                            /// k-th element in 
                                            /// the shared buffer
                                            cache[kdx + 0] = posp[jdx + 0];
                                            cache[kdx + 1] = posp[jdx + 1];
                                            cache[kdx + 2] = posp[jdx + 2];
                                            cache[kdx + 3] = posp[jdx + 3];
                                        }
                                        /// wait for all GPU threads 
                                        /// to complete the operation
                                        cpx_sync();

                                        /// compute i-th body interactions with 
                                        /// all bodies within the current 
                                        /// 256-body block
                                        for(int l = 0; l < 256; l++)
                                        {
                                            int jdx = 4 * l;
                            	
                                            float r[3];

                                            /// compute vector deltas
                                            r[0] = cache[jdx + 0] - 
                                                    posp[idx + 0];
                                            r[1] = cache[jdx + 1] - 
                                                    posp[idx + 1];
                                            r[2] = cache[jdx + 2] - 
                                                    posp[idx + 2];

                                            /// compute distance
                                            float dist = r[0] * r[0] + 
                                                         r[1] * r[1] + 
                                                         r[2] * r[2];
                                            dist += soften;

                                            /// 1 / dist^(3/2)
                                            float inv = rsqrtf(dist);
                                            float cube =  inv * inv * inv;

                                            /// compute force
                                            float s = cache[jdx + 3] * cube;

                                            /// superpose forces
                                            acc[0] += r[0] * s;
                                            acc[1] += r[1] * s;
                                            acc[2] += r[2] * s;
                                        }
                                        /// wait for all GPU threads 
                                        /// to complete their operation
                                        cpx_sync();
                                    }
                                    /// store forces
                                    forp[fdx + 0] = acc[0];
                                    forp[fdx + 1] = acc[1];
                                    forp[fdx + 2] = acc[2];
                                }
                            }
                            barrier();
                        }
                    }
                    barrier();
                }
                /// repeat infinitely until successful
                if(!error)
                {
                    break;
                }
            }
            barrier();
        }
        barrier();
    }
    /// sum up intermediate force vectors
    parallel()
    {
        for (int i = CPX_RANGE_START(bodies), 
               end = CPX_RANGE_END(bodies) ; i < end; i++)
        {
            int fdx = 3 * i;

            for(int j = 1; j < N / M; j++)
            {
                force[0][fdx + 0] += force[j][fdx + 0];
                force[0][fdx + 1] += force[j][fdx + 1];
                force[0][fdx + 2] += force[j][fdx + 2];
            }
        }
        barrier();
    }

    /// use parallel(1) to send everything to GPU if available
    parallel(N / M)
    {
        for(;;)
        {
            int start = CPX_RANGE_START(bodies);
            int len = CPX_RANGE_END(bodies) - start;

            /// access through pointers
            float* shared forp = &force[0][start * 3];
            float* shared velp = &vel[start * 4];
            float* shared posp = &pos[start * 4];
            int* shared lenp = &len;

            /// error-resilient programming facility
            int error = 0;
            /// access through pointer
            int* shared errp = &error;

            /// use parallel(1) to send everything to GPU if available
            parallel(1)
            {
                /// co-load all GPUs (and describe all dependencies)
                coload(
                       /// input forces
                       cpx_in   (float*, forp, len * sizeof(float) * 3),
                       /// input and output velocities
                       cpx_inout(float*, velp, len * sizeof(float) * 4),
                        /// input and output positions
                       cpx_inout(float*, posp, len * sizeof(float) * 4),

                       /// input constants
                       cpx_const(float, damp),
                       cpx_const(float, delta),
                       cpx_const(int, len),

                       /// error pointer
                       cpx_error(errp),
            
                       /// the actual number of iterations to perform
                       cpx_range(*lenp))
                {
                    /// loop over bodies (use built-in decomposition macros)
                    for (int i = CPX_RANGE_START(len), 
                           end = CPX_RANGE_END(len) ; i < end; ++i)
                    {
                        /// position, velocity and force indices
                        int idx = 4 * i;
                        int fdx = 3 * i;

                        /// read positions, velocities, forces, masses 
                        /// into local variables
                        float p[3], v[3], f[3];

                        p[0] = posp[idx + 0];
                        p[1] = posp[idx + 1];
                        p[2] = posp[idx + 2];

                        float mass = posp[idx + 3];
                    
                        v[0] = velp[idx + 0];
                        v[1] = velp[idx + 1];
                        v[2] = velp[idx + 2];

                        f[0] = forp[fdx + 0];
                        f[1] = forp[fdx + 1];
                        f[2] = forp[fdx + 2];

                        /// compute and superpose new velocities and positions

                        // acceleration = force / mass; 
                        // new velocity = old velocity + acceleration * 
                        //                delta time
                        v[0] += (f[0] * mass) * delta;
                        v[1] += (f[1] * mass) * delta;
                        v[2] += (f[2] * mass) * delta;

                        v[0] *= damp;
                        v[1] *= damp;
                        v[2] *= damp;

                        // new position = old position + velocity * delta time
                        p[0] += v[0] * delta;
                        p[1] += v[1] * delta;
                        p[2] += v[2] * delta;

                        /// store positions
                        posp[idx + 0] = p[0];
                        posp[idx + 1] = p[1];
                        posp[idx + 2] = p[2];

                        /// store velocities
                        velp[idx + 0] = v[0];
                        velp[idx + 1] = v[1];
                        velp[idx + 2] = v[2];
                    }
                }
                barrier();
            }
            /// repeat infinitely until successful
            if(!error)
            {
                break;
            }
        }
        barrier();
    }
}

void main()
{
    /// prime the coload functionality 
    parallel(1)
    {
        /// use empty coload operator to trigger lazy initialization
        coload(){}
        break();
    }

    /// initialize body system
    init_bodies(pos_par, vel_par, pscale, vscale, bodies);
    memset(force_par, 0, sizeof(force_par));

    /// time parallel version
    time = __rdtsc();

    /// update the system
    compute_bodies(step, pos_par, vel_par, force_par);

    time = __rdtsc() - time;

    printf("Time par: %I64d\n\n", time);

    /// print the results
    for (int i = 0; i < bodies; i++)
    {
        printf("H[%d]=%f\n", i, pos_par[i]);
    }
}
