//
//  File  : pi.cpp
//  (c)HOOPOE ProGroup 2013
//

//
// C= Programming Sample
//

/// Copyright string
char* copyright = "(c)HOOPOE ProGroup 2013";

#include <stdio.h>
#include <C=.h>

/// note that not all GPUs support doubles, so we use floats here
int main()
{
    /// the number of iterations to compute PI
    const int N = 0x1000000;
    /// the length of buffer to reduce intermediate values
    const int S = 0x100;
    /// reduction coefficient
    const int scale = N / S;

    /// the reduction buffer
    float buffer[S];
    /// reference through pointer as we access it from multiple CPUs
    float* volatile reduce = buffer;

    /// PI
    double pi = 0.;

    /// do on all CPUs
    parallel()
    {
        /// co-load all GPUs (and describe all dependencies)
        coload(
               /// describe the output array (type, pointer, size)
               cpx_out(float*, reduce, sizeof(buffer)),
               /// describe the reduction coefficient
               cpx_const(int, scale),
               /// pass N
               cpx_const(int, N),
               /// pass the actual number of iterations to perform
               cpx_range(S))
        {
            /// loop over S elements (use built-in decomposition macros)
            for(int i = CPX_RANGE_START(S), end = CPX_RANGE_END(S); 
                    i < end; i++)
            {
                float pi = 0;

                /// pre-compute PI for N / S elements
                for(int j = i * scale; j < (i + 1) * scale; j++)
                {
                    float t = (float)(j + 0.5) / N;

                    pi += 4.0 / (1.0 + t * t);
                }
                /// store PI in the output array
                reduce[i] = pi;
            }
        }
        barrier();
    }
    /// reduce the resulting S elements
    for(int i = 0; i < S; i++)
    {
        pi += reduce[i];
    }
    /// print PI
    printf("pi = %16.15f\n", pi / N);
}
